import pandas as _pd
import numpy as _np
import statsmodels.formula.api as _smf
import ctypes
from numpy.ctypeslib import ndpointer
import os
import multiprocessing as mp
import time
from numpy.random import default_rng, Generator, SFC64
import platform
from .extensions import csimOU as _csimOU
from .extensions import csimOUJ as _csimOUJ


class Result:
    def __init__(self):
        self.val = _pd.DataFrame()

    def update_result(self, val):
        self.val = _pd.concat([self.val, val], axis=1)  # append by columns


def is_iterable(x):
    try:
        iter(x)
        return True
    except:
        return False


def make_into_array(x, N):
    # make an array of same size as N+1
    if is_iterable(x):
        if len(x.shape) == 2:
            # if a 2D array is passed, return it as is
            # good for stocastic volatility matrix
            x = _np.vstack((x[0], x))
            return x

        if x.shape[0] == N:
            x = _np.append(x[0], x)
        else:
            raise ValueError(
                "if mu is passed as an iterable, it must be of length int(T/dt)"
            )

    else:
        x = _np.ones(N + 1) * x

    return x


def simGBM(s0=10, mu=0, sigma=0.2, r=0, T=1, dt=1 / 252, sims=1000, eps=None):
    """
    Simulates a Geometric Brownian Motion stochastic process (random walk)

    Parameters
    ----------
    s0 : float
        Starting value for GBM process at time = 0. By default, this is 10.
    mu : float
        Mean for normally distribution for normally distributed returns.
        Not used if eps is provided. By default, this is 0.
    sigma : float
        Annualized standard deviation for normally distributed returns. By default, this is 0.2.
        Not used if eps is provided.
    r : float
        Interest rate to discount the process by. By default, this is 0.
    T : float
        Time to maturity in years. But default, this is 1.
    dt : float
        Time step in period e.g. 1/252 = 1 business day. By default, this is 1/252.
    sims : int
        Number of simulations to run. By default, this is 1000.
    eps : numpy array
        Random numbers to use for the returns. If provided, mu and sigma are ignored.
        Must of size (p x sims) where p is the number of periods in T/dt.

    Returns
    -------
    A list of simulated values

    Examples
    --------
    >>> import risktools as rt
    >>> rt.simGBM(s0=5, mu=0, sigma=0.2, r=0.01, T=2, dt=1/252, sims=1000)
    """

    periods = int(T / dt)
    s = _np.zeros((periods + 1, sims))
    s = _pd.DataFrame(data=s)

    if eps is None:
        eps = _np.random.normal(mu, 1, size=(periods, sims))

    s.iloc[0, :] = s0
    s.iloc[1:, :] = eps

    # calc geometric brownian motion
    s.loc[1:, :] = _np.exp(
        (r - sigma**2 / 2) * dt + sigma * _np.sqrt(dt) * s.loc[1:, :]
    )

    s = s.cumprod()

    return s


def _import_csimOU():
    dir = os.path.dirname(os.path.realpath(__file__)) + "/../"  # + "/c/"

    ext = ".so"
    if platform.system() == "Windows":
        ext = ".dll"

    lib = ctypes.cdll.LoadLibrary(dir + "simOU" + ext)
    fun = lib.csimOU
    fun.restype = None
    fun.argtypes = [
        ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
        ctypes.c_double,
        ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
        ctypes.c_double,
        ctypes.c_double,
        ctypes.c_size_t,
        ctypes.c_size_t,
    ]
    fun.restype = ndpointer(ctypes.c_double, flags="C_CONTIGUOUS")
    return fun


def simOU(
    s0=5,
    mu=4,
    theta=2,
    sigma=1,
    T=1,
    dt=1 / 252,
    sims=1000,
    eps=None,
    seed=None,
    log_price=False,
    c=True,
):
    """
    Function for calculating an Ornstein-Uhlenbeck Mean Reversion stochastic process (random walk) with multiple
    simulations

    From Wikipedia:

    https://en.wikipedia.org/wiki/Ornstein-Uhlenbeck_process

    The process is a stationary Gauss–Markov process, which means that it is a Gaussian process, a Markov process,
    and is temporally homogeneous. In fact, it is the only nontrivial process that satisfies these three conditions,
    up to allowing linear transformations of the space and time variables. Over time, the process tends to drift
    towards its mean function: such a process is called mean-reverting.

    Parameters
    ----------
    s0 : float
        Starting value for mean reverting random walk at time = 0
    mu : float | array-like (1D or 2D)
        Mean that the stochastic diffusion equation will revert to. Scalars used for constants means through
        time and through various simulations. 1D arrays are supported for time varying means which must be
        the same length as T/dt (i.e. the number of periods). 2D arrays are also supported for stochastic
        mu where the first dimension is the number of periods and the second dimension equals the number of
        simulations.
    theta : float
        Mean reversion rate, higher number means it will revert slower
    sigma : float | array-like (1D or 2D)
        Annualized volatility or standard deviation. To calculate, take daily volatility and multiply by sqrt(T/dt).
        1D arrays are supported for time varying volatility which must be the same length as T/dt (i.e. the number of
        periods). 2D arrays are also supported for stochastic volatility where the first dimension is the number of
        periods and the second dimension equals the number of simulations.
    T : float or int
        Period length in years (i.e. 0.25 for 3 months)
    dt : float
        Time step size in fractions of a year. So a day would be 1/252, where 252 is the number of business
        days in a year
    sims : int
        Number of simulations to run. By default, this is 1000. Not used if eps is provided.
    eps : matrix-like[float]
        Random numbers to use for the returns. If provided, mu, sigma, T, dt and sims are ignored.
        Must of size (p x sims) where p is the number of periods in T, i.e. int(T/dt).
        Excludes time 0.
    seed : int
        To pass to numpy random number generator as seed. For testing only.
    log_price : bool
        Adds adjustment term to the mean reversion term if the prices passed are log prices. By
        default False.
    c : bool
        Whether or not to run C optimized code. By default True. Otherwise use python loop.

    Returns
    -------
    A pandas dataframe with the time steps as rows and the number of simulations as columns

    Examples
    --------
    >>> import risktools as rt
    >>> rt.simOU()
    """
    if eps is not None:
        sims = eps.shape[1]

    # number of business days in a year
    bdays_in_year = 252

    # calc periods
    N = int(T / dt)

    # print half-life of theta
    print("Half-life of theta in days = ", _np.log(2) / theta * bdays_in_year)

    # make mu array
    mu = make_into_array(mu, N)
    sigma = make_into_array(sigma, N)

    # make same size as 1D array of all periods and sims
    if len(mu.shape) == 1:
        mu = _np.tile(_np.array(mu), sims)
    else:
        mu = mu.flatten("F")

    # Don't run if 2D array passed for sigma
    if len(sigma.shape) == 1:
        sigma = _np.tile(_np.array(sigma), sims)
    else:
        sigma = sigma.flatten("F")

    if c == True:
        return _simOUc(
            s0=s0,
            mu=mu,
            theta=theta,
            T=T,
            dt=dt,
            sigma=sigma,
            sims=sims,
            eps=eps,
            seed=seed,
            log_price=log_price,
        )
    else:
        return _simOUpy(
            s0=s0,
            mu=mu,
            theta=theta,
            T=T,
            dt=dt,
            sigma=sigma,
            sims=sims,
            eps=eps,
            seed=seed,
            log_price=log_price,
        )


def _simOUc(s0, theta, mu, dt, sigma, T, sims=10, eps=None, seed=None, log_price=False):
    # calc periods
    N = int(T / dt)

    # check on random number size
    if (N + 1) * sims > 200_000_000:
        import warnings

        warnings.warn(
            """
            Note that this simulation will generate more than
            200M random numbers which may crash the python kernel. It may
            be a better idea to split the simulation into a series of smaller
            sims and then join them after.
            """
        )

    # generate a 1D array of random numbers that is based on a
    # 2D array of size P x S where P is the number of time steps
    # including s0 (so N + 1) and S is the number of sims. 1D
    # array will be of size P * S. This is actually the slowest
    # part.
    if eps is None:
        # rng = default_rng(seed)
        rng = Generator(SFC64(seed))
        x = rng.normal(loc=0, scale=1, size=((N + 1) * sims))
        x[0] = s0
        x[:: (N + 1)] = s0
    else:
        x = eps.T
        x = _np.c_[_np.ones(sims) * s0, x]
        x = x.reshape((N + 1) * sims)

    x = _csimOU(
        x, theta, mu, dt, sigma, rows=sims, cols=N + 1, log_price=int(log_price)
    )

    return _pd.DataFrame(x.reshape((sims, N + 1)).T)


def _simOUpy(
    s0, mu, theta, sigma, T, dt, sims=1000, eps=None, seed=None, log_price=False
):

    # number of periods dt in T
    N = int(T / dt)

    mu = mu.reshape((sims, N + 1)).T
    sigma = sigma.reshape((sims, N + 1)).T

    mu = _pd.DataFrame(mu)
    sigma = _pd.DataFrame(sigma)

    # init df with zeros, rows are steps forward in time, columns are simulations
    out = _np.zeros((N + 1, sims))
    out = _pd.DataFrame(data=out)

    # set first row as starting value of sim
    out.loc[0, :] = s0

    # calc gaussian vector
    if eps is None:
        # rng = default_rng(seed)
        rng = Generator(SFC64(seed))
        eps = rng.normal(size=(N, sims))

    out.iloc[1:, :] = eps

    if log_price:
        ss = 0.5 * sigma * sigma
        for i in range(1, N + 1):
            # calc step
            out.iloc[i, :] = (
                out.iloc[i - 1, :]
                + (theta * (mu.iloc[i, :] - out.iloc[i - 1, :]) - ss.iloc[i, :]) * dt
                + sigma.iloc[i, :] * out.iloc[i, :] * _np.sqrt(dt)
            )
    else:
        for i in range(1, N + 1):
            # calc step
            out.iloc[i, :] = (
                out.iloc[i - 1, :]
                + (theta * (mu.iloc[i, :] - out.iloc[i - 1, :])) * dt
                + sigma.iloc[i, :] * out.iloc[i, :] * _np.sqrt(dt)
            )

    return out


def simOUJ(
    s0=5,
    mu=5,
    theta=0.5,
    sigma=0.2,
    jump_prob=0.05,
    jump_avgsize=3,
    jump_stdv=0.05,
    T=1,
    dt=1 / 12,
    sims=1000,
    mr_lag=None,
    eps=None,
    elp=None,
    ejp=None,
    seed=None,
    c=True,
):
    """
    Function for calculating an Ornstein-Uhlenbeck Jump Mean Reversion stochastic process (random walk) with multiple
    simulations

    From Wikipedia:

    https://en.wikipedia.org/wiki/Ornstein-Uhlenbeck_process

    The process is a stationary Gauss–Markov process, which means that it is a Gaussian process, a Markov process,
    and is temporally homogeneous. In fact, it is the only nontrivial process that satisfies these three conditions,
    up to allowing linear transformations of the space and time variables. Over time, the process tends to drift
    towards its mean function: such a process is called mean-reverting.

    Parameters
    ----------
    s0 : float
        Starting value for mean reverting random walk at time = 0
    mu : float, int or pandas Series
        Mean that the function will revert to. Can be either a scalar value (i.e. 5) or a pandas series for a
        time dependent mean. If array-like, it must be the same length as T/dt (i.e. the number of periods)
    theta : float
        Mean reversion rate, higher number means it will revert slower
    sigma : float | array-like (1D or 2D)
        Annualized volatility or standard deviation. To calculate, take daily volatility and multiply by sqrt(T/dt).
        1D arrays are support for time varying volatility which must be the same length as T/dt (i.e. the number of
        periods). 2D arrays are also supported for stochastic volatility where the first dimension is the number of
        periods and the second dimension is the number of simulations.
    jump_prob : float
        Probablity of jumps for a Possion process.
    jump_avgsize : float
        Average size of jumps for a log normal distribution
    jump_stdv : float
        Standard deviation of average jump size for a log normal distribution
    T : float or int
        Period length in years (i.e. 0.25 for 3 months)
    dt : float
        Time step size in fractions of a year. So a day would be 1/252, where 252 is the number of business
        days in a year
    sims : int
        Number of simulations to run
    mr_lag : int, optional
        Lag in mean reversion. If None, then no lag is used. If > 0, then the diffusion does not immediately
        return the mean after a jump at theta but instead with remain near the jump level for mr_lag periods.
        By default, this is None.
    eps : numpy array, optional
        Array of random numbers to use for the simulation. If None, then random numbers are generated.
        By default, this is None.
    elp : numpy array, optional
        Array of random numbers to use for the log price jump. If None, then random numbers are generated.
        By default, this is None.
    ejp : numpy array, optional
        Array of random numbers to use for the jump size. If None, then random numbers are generated.
        By default, this is None.
    seed : int, optional
        To pass to numpy random number generator as seed. For testing only.
    log_price : bool, optional
        Adds adjustment term to the mean reversion term if the prices passed are log prices. By
        default False.
    c : bool, optional
        Whether or not to run C optimized code. By default True. Otherwise use python loop.

    Returns
    -------
    A pandas dataframe with the time steps as rows and the number of simulations as columns

    Examples
    --------
    >>> import risktools as rt
    >>> rt.simOUJ()
    """
    # number of business days in a year
    bdays_in_year = 252

    # number of periods dt in T
    N = int(T / dt)

    # check on random number size
    if (N + 1) * sims > 200_000_000:
        import warnings

        warnings.warn(
            """
            Note that this simulation will generate more than
            200M random numbers which may crash the python kernel. It may
            be a better idea to split the simulation into a series of smaller
            sims and then join them after.
            """
        )

    # print half-life of theta
    print("Half-life of theta in days = ", _np.log(2) / theta * bdays_in_year)

    if (eps is None) | (elp is None) | (ejp is None):
        rng = Generator(SFC64(seed))

    if eps is None:
        eps = rng.normal(size=(N, sims))
    else:
        N = eps.shape[0]
        sims = eps.shape[1]
    if elp is None:
        elp = rng.lognormal(mean=_np.log(jump_avgsize), sigma=jump_stdv, size=(N, sims))
    if ejp is None:
        ejp = rng.poisson(lam=jump_prob * dt, size=(N, sims))

    mu = make_into_array(mu, N)
    sigma = make_into_array(sigma, N)

    # repeats first row of eps, elp, and ejp to match the number of simulations
    eps = make_into_array(eps, N).astype(float)
    elp = make_into_array(elp, N).astype(float)
    ejp = make_into_array(ejp, N).astype(float)

    if c == True:
        if len(sigma.shape) == 2:
            sigma = sigma.T.reshape((N + 1) * sims)

        s = _simOUJc(
            s0,
            eps,
            elp,
            ejp,
            theta,
            mu,
            dt,
            sigma,
            sims,
            N,
            mr_lag,
            jump_prob,
            jump_avgsize,
        )
    else:
        # init df with zeros, rows are steps forward in time, columns are simulations
        s = _np.zeros((N + 1, sims))
        s = _pd.DataFrame(data=s)

        # set first row as starting value of sim
        s.iloc[0, :] = s0
        s = _simOUJpy(
            N, s, mu, theta, sigma, jump_prob, jump_avgsize, dt, mr_lag, eps, elp, ejp
        )

    return s


def _simOUJc(
    s0,
    eps,
    elp,
    ejp,
    theta,
    mu,
    dt,
    sigma,
    sims,
    N,
    mr_lag,
    jump_prob,
    jump_avgsize,
):

    # generate a 1D array of random numbers that is based on a
    # 2D array of size P x S where P is the number of time steps
    # including s0 (so N + 1) and S is the number of sims. 1D
    # array will be of size P * S. This is actually the slowest
    # part.
    eps[0, :] = s0

    # make same size as 1D array of all periods and sims
    eps = eps.T.reshape((N + 1) * sims)
    elp = elp.T.reshape((N + 1) * sims)
    ejp = ejp.T.reshape((N + 1) * sims)
    mu = _np.tile(_np.array(mu), sims)
    sigma = _np.tile(_np.array(sigma), sims)

    mr_lag = 0 if mr_lag is None else mr_lag

    x = _csimOUJ(
        x=eps,
        elp=elp,
        ejp=ejp,
        theta=theta,
        mu=mu,
        dt=dt,
        sigma=sigma,
        rows=sims,
        cols=N + 1,
        mr_lag=mr_lag,
        jump_prob=jump_prob,
        jump_avgsize=jump_avgsize,
    )

    return _pd.DataFrame(x.reshape((sims, N + 1)).T)


def _simOUJpy(
    N, s, mu, theta, sigma, jump_prob, jump_avgsize, dt, mr_lag, eps, elp, ejp
):

    mu = _np.vstack([mu] * s.shape[1]).T
    if len(sigma.shape) == 1:
        sigma = _np.vstack([sigma] * s.shape[1]).T

    mu = _pd.DataFrame(mu)
    sigma = _pd.DataFrame(sigma)
    eps = _pd.DataFrame(eps)
    elp = _pd.DataFrame(elp)
    ejp = _pd.DataFrame(ejp)

    for i in range(1, N + 1):
        # calc step
        # fmt: off
        s.iloc[i, :] = (
            s.iloc[i - 1, :]
            + theta
                * (mu.iloc[i, :] - jump_prob * jump_avgsize - s.iloc[i - 1, :])
                * s.iloc[i - 1, :]
                * dt
            + sigma.iloc[i, :] * s.iloc[i - 1, :] * eps.iloc[i, :] * _np.sqrt(dt)
            + ejp.iloc[i, :] * elp.iloc[i, :]
        )

        if mr_lag is not None:
            # if there is a jump in this step, add it to the mean reversion
            # level so that it doesn't drop back down to the given mean too
            # quickly. Simulates impact of lagged market response to a jump

            mu.iloc[(i):(i + mr_lag), :] += ejp.iloc[i, :] * elp.iloc[i, :]
            ejp.iloc[(i+1):(i + mr_lag), ejp.iloc[i,:] > 0] = 0 # stops double jumps

        # fmt: on

    return s


def fitOU(spread, dt=1 / 252, log_price=False, method="OLS", verbose=False):
    """
    Parameter estimation for the Ornstein-Uhlenbeck process

    Parameters
    ----------
    spread : array-like
        OU process as a list or series to estimate parameters for
    dt : float
        Time step size in fractions of a year. So a day would be 1/252,
        where 252 is the number of business days in a year. Default is 1/252.
        Only used if method is "OLS".
    log_price : bool
        If True, the spread is assumed to be log prices and the log of the spread is taken.
        Default is False.
    method : ['OLS', 'MLE']
        Method to use for parameter estimation. Default is 'OLS'.
    verbose : bool
        If True, prints the estimated parameters. Only used is method is OLS.
        Default is False.

    Returns
    -------
    Dictionary of alpha, mu and theta

    Examples
    --------
    >>> import risktools as rt
    >>> spread = rt.simOU(mu=5, theta=0.5, sigma=0.2, T=5, dt=1/252)
    >>> rt.fitOU(spread[0], method='MLE')
    """

    if log_price == True:
        spread = _np.log(spread)

    if method == "MLE":
        return _fitOU_MLE(spread)
    elif method == "OLS":
        return _fitOU_OLS(spread, dt, verbose)


def _fitOU_MLE(spread):
    """
    Parameter estimation for the Ornstein-Uhlenbeck process

    Parameters
    ----------
    spread : array-like
        OU process as a list or series to estimate parameters for

    Returns
    -------
    Dictionary of alpha, mu and theta

    Examples
    --------
    >>> import risktools as rt
    >>> spread = rt.simOU(mu=5, theta=0.5, signma=0.2, T=5, dt=1/250)
    >>> rt.fitOU(spread)
    """
    spread = _np.array(spread)
    n = len(spread)
    delta = 1

    Sx = spread[:-1].sum()
    Sy = spread[1:].sum()
    Sxx = (spread[:-1] ** 2).sum()
    Syy = (spread[1:] ** 2).sum()
    Sxy = (_np.multiply(spread[:-1], spread[1:])).sum()
    mu = (Sy * Sxx - Sx * Sxy) / ((n - 1) * (Sxx - Sxy) - (Sx**2 - Sx * Sy))

    theta = (
        -_np.log(
            (Sxy - mu * Sx - mu * Sy + (n - 1) * mu**2)
            / (Sxx - 2 * mu * Sx + (n - 1) * mu**2)
        )
        / delta
    )
    a = _np.exp(-theta * delta)
    sigmah2 = (
        Syy
        - 2 * a * Sxy
        + a**2 * Sxx
        - 2 * mu * (1 - a) * (Sy - a * Sx)
        + (n - 1) * mu**2 * (1 - a) ** 2
    ) / (n - 1)
    # print((sigmah2) * 2 * theta / (1 - a**2))
    sigma = _np.sqrt((sigmah2) * 2 * theta / (1 - a**2))
    theta = {"theta": theta, "mu": mu, "annualized_sigma": sigma}

    return theta


def _fitOU_OLS(spread, dt, verbose=False):
    """
    Parameter estimation for the Ornstein-Uhlenbeck process

    Parameters
    ----------
    spread : array-like
        OU process as a list or series to estimate parameters for
    dt : float
        Time step size in fractions of a year. So a day would be 1/252,
        where 252 is the number of business days in a year. Default is 1/252.
        Only used if method is "OLS".
    verbose : bool
        If True, prints the estimated parameters. Default is False.

    Returns
    -------
    Dictionary of annualized sigma, mu and theta

    Examples
    --------
    >>> import risktools as rt
    >>> spread = rt.simOU(mu=5, theta=0.5, signma=0.2, T=5, dt=1/250)
    >>> rt.fitOU(spread)
    """

    if isinstance(spread, _pd.DataFrame):
        raise ValueError("Spread must be a series or array-like, not a dataframe")

    df = _pd.DataFrame()

    df["mrg"] = spread
    df["delta"] = df["mrg"].diff().shift(-1)

    df = df.dropna().reset_index()

    mod = _smf.ols(formula="delta ~ mrg", data=df)
    res = mod.fit()

    theta = -res.params.mrg / dt
    mu = res.params.Intercept / theta / dt
    sigma = res.resid.std() / _np.sqrt(dt)

    if verbose == True:
        print(res.summary())
        print(
            "theta: ",
            round(theta, 2),
            "| mu: ",
            round(mu, 2),
            "| annualized_sigma: ",
            round(sigma, 2),
        )

    return {"theta": theta, "mu": mu, "annualized_sigma": sigma}


if __name__ == "__main__":
    import os
    import sys

    print(os.path.dirname(os.path.realpath(__file__)))

    eps = _pd.read_csv("./pytest/data/diffusion.csv", header=None)

    df = simGBM(
        s0=10, mu=0.0, sigma=0.2, r=0.05, T=1, dt=1 / 252, sims=20, eps=eps
    ).round(2)

    act = _pd.read_csv("./pytest/data/simGBM_output.csv")
    act = act.drop("t", axis=1).T.reset_index(drop=True).T.round(2)

    assert df.equals(act), "simGBM RTL eps failed"

    _np.random.seed(123)
    df = (
        simGBM(s0=10, mu=0.0, sigma=0.2, r=0.05, T=1, dt=1 / 252, sims=20)
        .astype("float")
        .round(4)
    )

    act = (
        _pd.read_csv("./pytest/data/simGBM_output_no_eps.csv").astype("float").round(4)
    )
    act = act.T.reset_index(drop=True).T

    assert df.equals(act), "simGBM generated eps failed"


def OU_lastcol(sims=100000, steps=250, T=25, sigma=0.1, mu=1, theta=1, S0=1):
    # https://stackoverflow.com/questions/24973961/vectorizing-an-equation
    dt = T / steps
    c = 1 - theta * dt
    cv = c ** _np.arange(steps)[::-1]

    R = _np.random.normal(theta * mu * dt, sigma, (sims, steps))

    SN = _np.dot(R, cv) + S0 * c**steps

    df = _pd.DataFrame(SN)
    return df


def stochastic_mu(mu, jump_prob, jump_size, dt, lag, N, sims, seed=None):
    """
    Generate a stochastic mu for the OU processes

    Parameters
    ----------
    mu : float
        The mean of the GBM process
    jump_prob : float
        The probability of a jump from a possion distribution
    jump_size : float
        The size of a jump in the mean
    jump_stdv : float
        The standard deviation of the jump size
    dt : float
        The time step size
    lag : int
        How long the mean reversion level will stay at the jump
        level for in time steps.
    N : int
        The number of time steps
    sims : int
        The number of simulations

    seed : int
        The seed for the random number generator. By default, the seed is None.

    Returns
    -------
    A dataframe of stochastic mu values of size N x sims where N is the number of time steps
    and sims is the number of simulations

    Examples
    --------
    >>> import risktools as rt
    >>> rt.stochastic_mu(mu=4, jump_prob=0.1, jump_avgsize=4, jump_stdv=0.5, dt=1/252, N=100, sims=10)
    """
    rng = Generator(SFC64(seed))
    # elp = rng.lognormal(mean=_np.log(jump_avgsize), sigma=jump_stdv, size=(N, sims))
    elp = _np.ones((N, sims)) * jump_size
    ejp = rng.poisson(lam=jump_prob * dt, size=(N, sims))

    ejp = _np.where(ejp == 0, _np.nan, ejp)
    ejp = _pd.DataFrame(ejp).fillna(method="ffill", axis=0, limit=lag).fillna(0).values

    mu_jump = mu * _np.ones((N, sims))
    mu_jump = mu_jump + elp * ejp

    return mu_jump



import warnings
import pandas as _pd
import numpy as _np
import quandl as _quandl
from . import data
import arch as _arch
from scipy.optimize import least_squares as _least_squares
from scipy import interpolate as _interpolate
import plotly.express as _px
import plotly.graph_objects as _go
import matplotlib.pyplot as _plt
from ._morningstar import *
from statsmodels.tsa.seasonal import STL as _STL
from pandas.plotting import (
    register_matplotlib_converters as _register_matplotlib_converters,
)
import seaborn as _sns

from ._pa import *


def ir_df_us(quandl_key=None, ir_sens=0.01, date=None):
    """
    Extracts US Tresury Zero Rates using Quandl

    Parameters
    ----------
    quandl_key : str
        Your Quandl key "yourkey" as a string, by default None. Optional as Quandl allows up to 50
        calls per day for free
    ir_sens : float
        Creates plus and minus IR sensitivity scenarios with specified shock value.
    date : string
        If date is not none, function will return ir curve for that date. By default None.

    Returns
    -------
    A pandas data frame of zero rates

    Examples
    --------
    >>> import risktools as rt
    >>> rt.ir_df_us()
    """

    if date is None:
        edt = _pd.Timestamp.now().floor("D")
    else:
        edt = _pd.to_datetime(date)
    # Last 30 days
    sdt = edt - _pd.DateOffset(days=30)

    if quandl_key is not None:
        _quandl.ApiConfig.api_key = quandl_key

    zero_1yr_plus = _quandl.get("FED/SVENY", start_date=sdt, end_date=edt)

    if zero_1yr_plus.index.max() < edt:
        warnings.warn(
            f"data only available until {zero_1yr_plus.index.max()}. Pulling data for {zero_1yr_plus.index.max()}"
        )
        edt = zero_1yr_plus.index.max()
        sdt = edt - _pd.DateOffset(days=30)

    fedsfund = _quandl.get("FED/RIFSPFF_N_D", start_date=sdt, end_date=edt).dropna()
    fedsfund["FedsFunds0"] = _np.log((1 + fedsfund.Value / 360) ** 365)
    fedsfund.drop("Value", axis=1, inplace=True)

    zero_tb = _quandl.get(
        ["FED/RIFLGFCM01_N_B", "FED/RIFLGFCM03_N_B", "FED/RIFLGFCM06_N_B"],
        start_date=sdt,
        end_date=edt,
    ).dropna()
    zero_tb.columns = zero_tb.columns.str.replace(" - Value", "", regex=False)

    # get most recent full curve (some more recent days will have NA columns)
    x = fedsfund.join(zero_tb).join(zero_1yr_plus).dropna().iloc[-1, :].reset_index()
    x.columns = ["maturity", "yield"]
    x["index"] = x["maturity"]
    x["yield"] /= 100
    x["maturity"] = x.maturity.str.extract("(\d+)").astype("float")

    # change maturity numbers to year fraction for first four rows
    x.iloc[1:4, x.columns.get_loc("maturity")] /= 12.0
    x.iloc[0, x.columns.get_loc("maturity")] = 1 / 365
    # x.maturity[1:4] /= 12
    # x.maturity[0] = 1/365

    # add new row for today, same yield as tomorrow
    x = _pd.concat(
        [_pd.DataFrame({"maturity": [0], "yield": [x["yield"][0]]}), x],
        ignore_index=True,
    )

    x["discountfactor"] = _np.exp(-x["yield"] * x.maturity)
    x["discountfactor_plus"] = _np.exp(-(x["yield"] + ir_sens) * x.maturity)
    x["discountfactor_minus"] = _np.exp(-(x["yield"] - ir_sens) * x.maturity)

    x.loc[0, "index"] = "0"
    x = x.set_index("index")

    x = x[
        [
            "yield",
            "maturity",
            "discountfactor",
            "discountfactor_plus",
            "discountfactor_minus",
        ]
    ]

    return x


def bond(ytm=0.05, c=0.05, T=1, m=2, output="price"):
    """
    Compute bond price, cash flow table and duration

    Parameters
    ----------
    ytm : float
        Yield to Maturity
    c : float
        Coupon rate per annum
    T : float
        Time to maturity in years
    m : float
        Periods per year for coupon payments e.g semi-annual = 2.
    output : string
        "price", "df" or "duration", by default "df"

    Returns
    -------
    Price scalar, cash flows data frame and/or duration scalar

    Examples
    --------
    >>> import risktools as rt
    >>> rt.bond(ytm = 0.05, c = 0.05, T = 1, m = 2, output = "price")
    >>> rt.bond(ytm = 0.05, c = 0.05, T = 1, m = 2, output = "df")
    >>> rt.bond(ytm = 0.05, c = 0.05, T = 1, m = 2, output = "duration")
    """
    assert output in [
        "df",
        "price",
        "duration",
    ], "output not a member of ['df','price','duration']"

    # init df
    df = _pd.DataFrame(
        {"t_years": _np.arange(1 / m, T + 1 / m, 1 / m), "cf": [c * 100 / m] * (T * m)}
    )
    df["t_periods"] = df.t_years * m
    df.loc[df.t_years == T, "cf"] = c * 100 / m + 100
    df["disc_factor"] = 1 / ((1 + ytm / m) ** df["t_periods"])
    df["pv"] = df.cf * df.disc_factor
    price = df.pv.sum()
    df["duration"] = df.pv * df.t_years / price

    if output == "price":
        ret = price
    elif output == "df":
        ret = df
    elif output == "duration":
        ret = df.duration.sum()
    else:
        ret = None

    return ret


def trade_stats(R, Rf=0):
    """
    Compute list of risk reward metrics

    Parameters
    ----------
    R : Series or DataFrame
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    Rf : float
        risk free rate, in same period as your returns, or as a single
        digit average

    Returns
    -------
    Dictionary with cummulative returns, annual returns, Annualized Sharpe ratio, Omega Sharpe ratio,
    Win &, % in the market, and drawdown specs

    Examples
    --------
    >>> from pandas_datareader import data
    >>> df = data.DataReader(["SPY","AAPL"],  "yahoo", '2000-01-01', '2012-01-01')
    >>> df = df.pct_change()
    >>> df = df.asfreq('B')
    >>> rt.trade_stats(df[('Adj Close','SPY')])
    """

    # r = R.dropna() # don't use, messes up freq check
    r = R.copy()

    rs = dict()

    # convert series into dataframe for flexibility
    series_flag = False
    if isinstance(r, _pd.Series):
        r = _pd.DataFrame({"trade_stats": r})
        series_flag = True

    for lab, con in r.items():
        rs[lab] = dict()
        y = find_drawdowns(con)
        rs[lab]["cum_ret"] = return_cumulative(con, geometric=True)
        rs[lab]["ret_ann"] = return_annualized(con, scale=252)
        rs[lab]["sd_ann"] = sd_annualized(con, scale=252)
        rs[lab]["omega"] = omega_sharpe_ratio(con, MAR=Rf)  # * _np.sqrt(252)
        rs[lab]["sharpe"] = sharpe_ratio_annualized(con, Rf=Rf)

        # need to dropna to calc perc_win properly
        con_clean = con.dropna()
        rs[lab]["perc_win"] = (
            con_clean[con_clean > 0].shape[0] / con_clean[con_clean != 0].shape[0]
        )
        rs[lab]["perc_in_mkt"] = con_clean[con_clean != 0].shape[0] / con_clean.shape[0]

        rs[lab]["dd_length"] = max(y["length"])
        rs[lab]["dd_max"] = min(y["return"])

    if series_flag == True:
        rs = rs["trade_stats"]

    return rs


def returns(df, ret_type="abs", period_return=1, spread=False):
    """
    Computes periodic returns from a dataframe ordered by date

    Parameters
    ----------
    df : Series or DataFrame
        If pandas series, series must have a single datetime index or a multi-index in the same of ('asset name','date'). Values must be asset prices
        If pandas dataframe, dataframe must have a single datetime index with asset prices as columns
    ret_type : str
        "abs" for absolute, "rel" for relative, or "log" for log returns. By default "abs"
    period_return : int
        Number of rows over which to compute returns. By default 1
    spread : bool
        True if you want to spread into a wide dataframe. By default False

    Returns
    -------
    A pandas series of returns.

    Examples
    --------
    >>> import risktools as rt
    >>> rt.returns(df = rt.data.open_data('dflong'), ret_type = "rel", period_return = 1, spread = True)
    >>> rt.returns(df = rt.data.open_data('dflong'), ret_type = "rel", period_return = 1, spread = False)
    """
    df = _check_df(df)

    if isinstance(df, _pd.Series):
        df = _pd.DataFrame({df.name: df})
    elif isinstance(df, _pd.DataFrame) == False:
        raise ValueError("df is not a pandas Series or DataFrame")

    index_flag = False
    if len(df.index.names) == 1:
        # convert single index into multi-index by adding dummy index in position 0 to allow for groupby
        df["dummy"] = "dummy"
        df = df.set_index("dummy", append=True)
        df = df.swaplevel()
        index_flag = True

    if (spread == True) & (len(df.index.names) == 1):
        raise ValueError(
            "You have set spread to True but only passed a single index, not a multi-index. There is nothing to spread"
        )

    # df = df.dropna()

    # calc return type
    if ret_type == "abs":
        df = df.groupby(level=0, group_keys=False).apply(lambda x: x.diff())
    elif ret_type == "rel":
        df = df.groupby(level=0, group_keys=False).apply(lambda x: x / x.shift(period_return) - 1)
    elif ret_type == "log":
        if df[df < 0].count().sum() > 0:
            warnings.warn(
                "Negative values passed to log returns. You will likely get NaN values using log returns",
                RuntimeWarning,
            )
        df = df.groupby(level=0, group_keys=False).apply(lambda x: _np.log(x / x.shift(period_return)))
    else:
        raise ValueError("ret_type is not valid")

    if spread:
        # return spread df, don't dropna or it will remove entire rows
        df = df.unstack(level=0).droplevel(level=0, axis=1)
        return df
    else:
        if index_flag == True:
            # drop dummy index to return single index
            df = df.droplevel(0)
        if len(df.columns) == 1:
            return df.dropna().iloc[:, 0]
        else:
            return df.dropna()


def roll_adjust(
    df, commodity_name="cmewti", roll_type="Last_Trade", roll_sch=None, *args
):
    """
    Returns a pandas series adjusted for contract roll. The methodology used to adjust returns is to remove the daily returns on
    the day after expiry and for prices to adjust historical rolling front month contracts by the size of the roll at
    each expiry. This is conducive to quantitative trading strategies as it reflects the PL of a financial trader.

    Note that this will apply a single expiry schedule to all assets passed to this function. To apply different roll schedules
    by asset type, perform a separate function call.

    Parameters
    ----------
    df : Series | DataFrame
        pandas series or dataframe with a with a datetime index and which columns are asset prices or with a dataframe with a multi-index in the shape ('asset name','date')
    commodity_name : str
        Name of commodity in expiry_table. See example below for values.
    roll_type : str
        Type of contract roll: "Last_Trade" or "First_Notice". By default "Last_Trade"
    roll_sch : Series
        For future capability. Optional
    *args: Other parms to pass to function

    Returns
    -------
    Roll-adjusted pandas series object of returns with datetime index or multi-index with the shape ('asset name','date')

    Examples
    --------
    >>> import risktools as rt
    >>> dflong = rt.data.open_data('dflong')
    >>> rt.data.open_data('expiry_table').cmdty.unique() # for list of commodity names
    >>> ret = rt.returns(df=dflong, ret_type="abs", period_return=1, spread=True)
    >>> ret = ret.iloc[:,0]
    >>> rt.roll_adjust(df=ret, commodity_name="cmewti", roll_type="Last_Trade")
    """

    df = _check_df(df)

    if isinstance(df, _pd.Series):
        df = _pd.DataFrame({df.name: df})
    elif isinstance(df, _pd.DataFrame) == False:
        raise ValueError("df is not a pandas Series or DataFrame")

    if roll_sch is None:
        roll_sch = data.open_data("expiry_table")
        roll_sch = roll_sch[roll_sch.cmdty == commodity_name]
        roll_sch = roll_sch[roll_type]

    df["expiry"] = df.index.isin(roll_sch, level=-1)
    df["expiry"] = df["expiry"].shift()

    df = df[df.expiry == False].drop("expiry", axis=1)

    if len(df.columns) == 1:
        return df.dropna().iloc[:, 0]
    else:
        return df.dropna()


def garch(df, out="data", scale=None, show_fig=True, forecast_horizon=1, **kwargs):
    """
    Computes annualised Garch(1,0,1) volatilities using arch package. Note that the
    RTL package uses a sGarch model, but there is no similiar implementation in
    python.

    Parameters
    ----------
    df : DataFrame
        Wide dataframe with date column and single series (univariate).
    out : str
        "plotly" or "matplotlib" to return respective chart types. "data" to return data or "fit" for garch fit output. By default 'data'
    show_fig : bool
        Only used if out is 'matplotlib' or 'plotly'. If True, function will display plots as well as return their respective fig object
    **kwargs
        key-word parameters to pass tp arch_model. if none, sets p=1, o=0, q=1

    Returns
    -------
    out='data' returns a pandas df
    out='fit' returns an arch_model object from the package arch
    out='matplotlib' returns matplotlib figure and axes objects and displays the plot
    out='plotly' returns a plotly figure object and displays the plot

    Examples
    --------
    >>> import risktools as rt
    >>> dflong = rt.data.open_data('dflong')
    >>> dflong = dflong['CL01']
    >>> df = rt.returns(df=dflong, ret_type="rel", period_return=1)
    >>> df = rt.roll_adjust(df=df, commodity_name="cmewti", roll_type="Last_Trade")
    >>> rt.garch(df, scale=252, out="data")
    >>> rt.garch(df, scale=252, out="fit")
    >>> rt.garch(df, scale=252, out="plotly")
    >>> rt.garch(df, scale=252, out="matplotlib")
    """
    df = _check_df(df).sort_index()

    # if kwargs is None:
    #     kwargs = {"p": 1, "o": 0, "q": 1}

    df = df - df.mean()

    # find a more robust way to do this, rolladjust may break it
    freq = _pd.infer_freq(df.index[-10:])

    if scale is None:
        if freq is None:
            raise ValueError(
                "Could not infer frequency of timeseries, please provide scale parameter instead"
            )
        elif (freq == "B") or (freq == "D"):
            scale = 252
        elif freq[0] == "W":
            scale = 52
        elif (freq == "M") or (freq == "MS"):
            scale = 12
        elif (freq == "Q") or (freq == "QS"):
            scale = 4
        elif (freq == "Y") or (freq == "YS"):
            scale = 1
        else:
            raise ValueError(
                "Could not infer frequency of timeseries, please provide scale parameter instead"
            )

    # a standard GARCH(1,1) model
    garch = _arch.arch_model(df, **kwargs)
    garch_fitted = garch.fit()

    # # calc annualized volatility from variance
    yhat = _np.sqrt(
        garch_fitted.forecast(horizon=forecast_horizon, start=0, reindex=True).variance
    ) * _np.sqrt(scale)

    if out == "data":
        return yhat
    elif out == "fit":
        return garch_fitted
    elif out == "plotly":
        fig = _px.line(yhat)
        fig.show()
        return fig
    elif out == "matplotlib":
        fig, ax = _plt.subplots(1, 1)
        yhat.plot(ax=ax)
        fig.show()
        return fig, ax


def prompt_beta(df, period="all", beta_type="all", output="chart"):
    """
    Returns array/dataframe of betas for futures contract returns of a commodity
    with it's front contract (i.e. next most expirying contract). For use with futures
    contracts (i.e. NYMEX WTI: CL01, CL02, CL03 and so forth) with standardized expiry periods.

    Using the WTI example, the betas will represent the covariance of daily returns of
    CL02, CL03, CL04 with CL01. The covariance of CL01 with CL01 is also returned, but is
    always 1.

    This function uses the single factor model (CAPM) beta. See the function CAPM_beta
    for more details

    Parameters
    ----------
    df : DataFrame
        Wide dataframe with datetime index and multiple series columns for each futures contract.
        Always use continuous contracts for columns.
    period : [str | int | float]
        Timeframe to use to calculate beta. "all" to use all data available or scalar number
        n to only use the last n periods/rows of data from the last, by default 'all'. i.e.
        for WTI contracts (CL), 30 would be the last 30 days. Recommend running roll_adjust
        function prior to using prompt_beta to remove swings from contract expiry
    beta_type : {'all', 'bull', 'bear'}, default 'all'
        Beta types to return
    output : {'betas', 'chart', 'stats'}, default 'chart'
        Output type

    Returns
    -------
    chart
        A plotly figure with the beta lines charted for 'all', 'bear' and 'bull' markets
    betas
        A dataframe of betas by contract order for 'all', 'bear' and 'bull' markets
    stats
        A scipy object from a least_squares fit of the betas by market type. Model used
        to fit betas was of the form:
        .. math:: \{beta} = x0 * exp(x1*t) + x2
        where t is the contract order (1, 2, 3 etc..., lower for expirying sooner)

    chart, df of betas or stats

    Examples
    --------
    >>> import risktools as rt
    >>> dfwide = rt.data.open_data('dfwide')
    >>> col_mask = dfwide.columns[dfwide.columns.str.contains('CL')]
    >>> dfwide = dfwide[col_mask]
    >>> x = rt.returns(df=dfwide, ret_type="abs", period_return=1)
    >>> x = rt.roll_adjust(df=x, commodity_name="cmewti", roll_type="Last_Trade")
    >>> rt.prompt_beta(df=x, period="all", beta_type="all", output="chart")
    >>> rt.prompt_beta(df=x, period="all", beta_type="all", output="betas")
    >>> rt.prompt_beta(df=x, period="all", beta_type="all", output="stats")
    >>> rt.prompt_beta(df=x, period=30, beta_type="all", output="plotly")
    """
    df = df.copy()
    # this assumes that the numeric component of the column name represents
    # an order to the asset contract
    term = df.columns.str.replace("[^0-9]", "", regex=True).astype(int)

    if isinstance(period, (int, float)):
        df = df.sort_index()
        df = df.iloc[-period:]

    # leave only contract # in column names
    df.columns = term
    df = df.sort_index()

    # calculate betas by market type (mkt is all types) using front contract as the benchmark
    mkt = CAPM_beta(df, df.iloc[:, 0])
    bull = CAPM_beta(df, df.iloc[:, 0], kind="bull")
    bear = CAPM_beta(df, df.iloc[:, 0], kind="bear")

    # create array for non-linear least squares exponential
    prompt = _np.arange(0, mkt.shape[0]) + 1

    # proposed model for beta as a function of prompt
    def beta_model(x, prompt):
        return x[0] * _np.exp(x[1] * prompt) + x[2]

    # cost function for residuals. Final equation that we're trying to minimize is
    # beta - x0*exp(x1*prompt) + x2
    def beta_residuals(x, beta, prompt):
        return beta - beta_model(x, prompt)

    # run least squares fit. Note that I ignore the first row of mkt and prompt arrays since
    # correlation of a var with itself should always be 1. Also, the beta of the second contract
    # will likely be a lot less then 1, and so ignoring the 1st contract will allow for a better fit
    r = _least_squares(
        beta_residuals, x0=[-1, -1, -1], args=(_np.array(mkt.iloc[1:]), prompt[1:])
    )

    # construct output df
    out = _pd.DataFrame()
    out["all"] = mkt
    out["bull"] = bull
    out["bear"] = bear

    if output == "stats":
        return r
    elif output == "betas":
        return out
    elif output == "chart":
        out = out[
            1:
        ]  # exclude beta of front contract with itself (always 1, no information)
        fig = _go.Figure()
        fig.add_trace(_go.Scatter(x=out.index, y=out["all"], mode="lines", name="all"))
        fig.add_trace(
            _go.Scatter(x=out.index, y=out["bear"], mode="lines", name="bear")
        )
        fig.add_trace(
            _go.Scatter(x=out.index, y=out["bull"], mode="lines", name="bear")
        )
        fig.update_xaxes(range=[out.index.min() - 1, out.index.max() + 1])
        fig.update_layout(
            title="Contract Betas vs Front Contract: Bear (Bull) = Beta in Down (Up) Moves",
            xaxis_title="Contract",
            yaxis_title="Beta",
            legend_title="Market Type",
        )
        return fig


def npv(
    init_cost=-375,
    C=50,
    cf_freq=0.25,
    F=250,
    T=2,
    disc_factors=None,
    break_even=False,
    be_yield=0.01,
):
    """
    Compute Net Present Value using discount factors from ir_df_us()

    Parameters
    ----------
    init_cost : float | int
        Initial investment cost
    C : float | int
        Periodic cash flow
    cf_freq : float
        Cash flow frequency in year fraction e.g. quarterly = 0.25
    F : float | int
        Final terminal value
    T : float | int
        Final maturity in years
    disc_factors : DataFrame
        Data frame of discount factors using ir_df_us() function.
    break_even : bool
        True when using a flat discount rate assumption.
    be_yield : float | int
        Set the flat IR rate when beak_even = True.

    Returns
    -------
    A python dictionary with elements 'df' as dataframe and 'npv' value as a float

    Examples
    --------
    >>> import risktools as rt
    >>> ir = rt.ir_df_us(ir_sens=0.01)
    >>> rt.npv(init_cost=-375, C=50, cf_freq=0.5, F=250, T=2, disc_factors=ir, break_even=True, be_yield=.0399)
    """
    disc_factors = disc_factors.copy()

    if disc_factors is None:
        raise ValueError(
            "Please input a discount factor dataframe into disc_factors (use ir_df_us to get dataframe)"
        )

    if break_even == True:
        disc_factors["yield"] = be_yield
        disc_factors["discountfactor"] = _np.exp(
            -disc_factors["yield"] * disc_factors.maturity
        )

    n = len(_np.arange(0, T, cf_freq)) + 1

    disc_intp = _interpolate.splrep(disc_factors.maturity, disc_factors.discountfactor)

    df = _pd.DataFrame(
        {
            "t": _np.append(
                _np.arange(0, T, cf_freq), [T]
            ),  # need to append T since _np.arange is of type [a,b)
            "cf": _np.ones(n) * C,
        }
    )

    df.loc[df.t == 0, "cf"] = init_cost
    df.loc[df.t == T, "cf"] = F

    df["df"] = _interpolate.splev(df.t, disc_intp)
    df["pv"] = df.cf * df.df

    return df


def crr_euro(s=100, x=100, sigma=0.2, Rf=0.1, T=1, n=5, type="call"):
    """
    European option binomial model on a stock without dividends. For academic purposes only.
    Use fOptions::CRRBinomialTreeOptions for real-life usage in R or Python equivalent.

    Parameters
    ----------
    s : float
        Stock price, by default 100
    x : float
        Strike price, by default 100
    sigma : float
        Implied volatility, by default 0.20
    Rf : float
        Risk-free rate, by default 0.1
    T : float
        Time to maturity in years, by default 1
    n :
        Number of time steps, by default 5. Internally dt = T/n.
    type : str
        "call" or "put", by default "call"

    Returns
    -------
    List of asset price tree, option value tree and option price.

    Examples
    --------
    >>> import risktools as rt
    >>> rt.crr_euro(s=100, x=100, sigma=0.2, Rf=0.1, T=1, n=5, type="call")
    """

    dt = T / n

    # define u, d, and risk-neutral probability
    u = _np.exp(sigma * _np.sqrt(dt))
    d = _np.exp(-sigma * _np.sqrt(dt))
    q = (_np.exp(Rf * dt) - d) / (u - d)

    # define our asset tree prices
    asset = _np.zeros([n + 1, n + 1])

    for i in range(0, n + 1):
        for j in range(0, i + 1):
            asset[i, j] = s * (u ** j) * (d ** (i - j))

    # create matrix of the same dims as asset price tree
    option = _np.zeros([n + 1, n + 1])
    # replace last row with maturity payoffs
    if type == "call":
        option[-1, :] = asset[-1, :] - x
    elif type == "put":
        option[-1, :] = x - asset[-1, :]
    else:
        raise ValueError("type is not of ('call','put')")
    option[option < 0] = 0

    # we discount recursively starting from final payoff (last row of tree)
    # starting at the second last period based on final payoffs
    for i in range(n - 1, -1, -1):  # loop for n-1 to 0 because range(a,b) is [a,b)
        for j in range(0, i + 1):
            option[i, j] = (
                (1 - q) * option[i + 1, j] + q * option[i + 1, j + 1]
            ) / _np.exp(Rf * dt)

    # indicator if model can be used sigma > rsqrt(dt)
    if sigma > _np.sqrt(dt) * Rf:
        note = "ok"
    else:
        note = "sigma < Rf*sqrt(dt) do not use"

    return {"asset": asset, "option": option, "price": option[0, 0], "note": note}


def stl_decomposition(
    df,
    output="chart",
    period=None,
    seasonal=7,
    seasonal_deg=1,
    resample_freq="M",
    **kwargs,
):
    """
    Provides a summary of returns distribution using the statsmodels.tsa.seasonal.STL class.
    Resamples all data to  prior to STL if

    Parameters
    ----------
    df : Series
        A pandas series (univariate) with a datetime index.
    output : str
        "chart" to see output as a graph, "data" for results as a list. By default "chart"
    seasonal : int, optional
        Length of the seasonal smoother. Must be an odd integer, and should
        normally be >= 7 (default).
    seasonal_deg : int, optional
        Degree of seasonal LOESS. 0 (constant) or 1 (constant and trend). By default 1
    resample_freq : str
        Resampling frequency to use in pandas resample function. Set to None to not resample
    **kwargs :
        Other parms



    season - The length of the seasonal smoother. Must be odd.

    trend - The length of the trend smoother, usually around 150% of season. Must be odd and larger than season.

    low_pass - The length of the low-pass estimation window, usually the smallest odd number larger than the periodicity of the data.


    Returns
    -------
    a chart or statsmodels.tsa.seasonal.DecomposeResult object of results

    Examples
    --------
    >>> import risktools as rt
    >>> df = rt.data.open_data('dflong')
    >>> df = df['CL01']
    >>> rt.stl_decomposition(df, output="chart", seasonal=13, seasonal_deg=1, resample_freq='M')
    >>> rt.stl_decomposition(df, output="data", seasonal=13, seasonal_deg=1, resample_freq='M')
    """

    _register_matplotlib_converters()
    _sns.set_style("darkgrid")

    if resample_freq is not None:
        df = df.resample(resample_freq).mean()

    stl = _STL(
        df,
        period=period,
        seasonal=seasonal,
        seasonal_deg=seasonal_deg,
        robust=True,
        **kwargs,
    )
    res = stl.fit()

    if output == "chart":
        fig = res.plot()
        fig.set_size_inches(15, 15)
        return fig  # res.plot()
    else:
        return res


def get_eia_df(tables, key, version=2):
    """
    Function for download data from the US Government EIA and return it as a pandas dataframe/series

    To get an api key, go to https://www.eia.gov/opendata/register.php
    To get table names, search via API explorer at https://www.eia.gov/opendata/qb.php

    Parameters
    ----------
    tables : List[Tuple[str]]
        EIA series to return. Can be a list or tuple of tables as well.
    key : str
        EIA key.
    version : 1 | 2
        API version to use, can be either 1 or 2. By default 2. As of Nov 2022, EIA is no
        longer support v1 of the API. 

    Returns
    -------
    pandas dataframe or series depending on the number of tables requested

    Examples
    --------
    >>> import risktools as rt
    >>> rt.get_eia_df('PET.WDIIM_R20-Z00_2.W', key=eia_key)
    """

    if int(version) == 1:
        return _get_eia_df_v1(tables, key)
    else:
        return _get_eia_df_v2(tables, key)


def _get_eia_df_v1(tables, key):
    """
    Function for download data from the US Government EIA and return it as a pandas dataframe/series
    API version 1.

    To get an api key, go to https://www.eia.gov/opendata/register.php
    To get table names, search via API explorer at https://www.eia.gov/opendata/qb.php

    Parameters
    ----------
    tables : List[Tuple[str]]
        EIA series to return. Can be a list or tuple of tables as well.
    key : str
        EIA key.

    Returns
    -------
    pandas dataframe or series depending on the number of tables requested

    Examples
    --------
    >>> import risktools as rt
    >>> rt.get_eia_df('PET.WDIIM_R20-Z00_2.W', key=eia_key)
    """
    import requests
    import json

    if isinstance(tables, list) == False:
        tables = [tables]

    eia = _pd.DataFrame()

    for tbl in tables:
        url = r"http://api.eia.gov/series/?api_key={}&series_id={}&out=json".format(
            key, tbl
        )
        tmp = json.loads(requests.get(url).text)

        tf = _pd.DataFrame(tmp["series"][0]["data"], columns=["date", "value"])
        tf["table_name"] = tmp["series"][0]["name"]
        tf["series_id"] = tmp["series"][0]["series_id"]
        eia = _pd.concat([eia, tf], axis=0)
        # eia = eia.append(tf)

    eia.loc[eia.date.str.len() < 7, "date"] += "01"

    eia.date = _pd.to_datetime(eia.date)
    return eia


def _get_eia_df_v2(tables, key):
    """
    Function for download data from the US Government EIA and return it as a pandas dataframe/series.
    API version 2.

    To get an api key, go to https://www.eia.gov/opendata/register.php
    To get table names, search via API explorer at https://www.eia.gov/opendata/qb.php

    Parameters
    ----------
    tables : List[Tuple[str]]
        EIA series to return. Can be a list or tuple of tables as well.
    key : str
        EIA key.

    Returns
    -------
    pandas dataframe or series depending on the number of tables requested

    Examples
    --------
    >>> import risktools as rt
    >>> rt.get_eia_df_v2('PET.WDIIM_R20-Z00_2.W', key=eia_key)
    """
    import requests
    import json

    if isinstance(tables, list) == False:
        tables = [tables]

    eia = _pd.DataFrame()

    for tbl in tables:
        url = f"http://api.eia.gov/v2/seriesid/{tbl}?api_key={key}"
        tmp = json.loads(requests.get(url).text)

        tf = _pd.DataFrame(tmp['response']['data'], columns=["period","series-description", "value"])
        tf["series_id"] = tbl
        eia = _pd.concat([eia, tf], axis=0)
        # eia = eia.append(tf)
    
    eia = eia.rename(columns={'period':'date','series-description':'table_name'})
    eia.loc[eia.date.str.len() < 7, "date"] += "01"
    eia.date = _pd.to_datetime(eia.date)
    return eia[['date','value','table_name','series_id']]


def _check_df(df):
    # if isinstance(df.index, _pd.DatetimeIndex):
    #     # reset index if the df index is a datetime object
    #     df = df.reset_index().copy()

    return df.copy()


def infer_freq(x, multiplier=False):
    """
    Function to infer the frequency of a time series. Improvement over 
    pandas.infer_freq as it can handle missing days/holidays. Note that
    for business days it will return 'D' vs 'B'

    Parameters
    ----------
    x : DataFrame | Series
        Time series. Index MUST be a datetime index.
    multiplier : bool
        If True, returns annualization factor for financial time series:
        252 for daily/business day
        52 for weekly
        12 for monthly
        4 for quarterly
        1 for annual
    Returns
    -------
    str is multiplier = False
    """

    # searches for 3 consecutive rows and then infers freq
    x = x.copy()
    x.index = _pd.to_datetime(x.index)
    diffs = x.index[1:] - x.index[:-1]
    min_delta = diffs.min()
    mask = (diffs == min_delta)[:-1] & (diffs[:-1] == diffs[1:])
    pos = _np.where(mask)[0][0]

    freq = _pd.infer_freq(x.index[pos : pos + 3])
    if multiplier == False:
        return freq
    else:
        if freq in ["D", "B"]:
            return 252
        elif freq[0] == "W":
            return 52
        elif freq[0] == "M":
            return 12
        elif freq[0] == "Q":
            return 4
        elif freq[0] == "A":
            return 1




import setuptools
from setuptools import Extension
import os
import sysconfig
from Cython.Distutils import build_ext
import Cython.Build
import platform
import numpy


class NoSuffixBuilder(build_ext):
    def get_ext_filename(self, ext_name):
        filename = super().get_ext_filename(ext_name)
        suffix = sysconfig.get_config_var("EXT_SUFFIX")
        ext = os.path.splitext(filename)[1]
        return filename.replace(suffix, "") + ext


extensions = [
    Extension(
        name="extensions",
        sources=["src/risktools/pyx/sims.pyx"],
        include_dirs=[numpy.get_include()]
        # extra_compile_args=['-fPIC', '-shared']
    )
]

requirements = [
    "pandas",
    "numpy",
    "numba",  # requried to install arch on Windows
    "matplotlib",
    "plotly",
    "quandl",
    "scikit-learn",
    "arch",
    "scipy",
    "statsmodels",
    "seaborn",
    "pandas_datareader",
]

preqs = ">=3.7"

if platform.system() == "Windows":
    requirements.remove("numba")

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="risktools",
    version="0.2.8.4",
    author="Ben Cho",
    license="gpl-3.0",  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    author_email="ben.cho@gmail.com",
    description="Python implementation of the R package RTL",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    package_data={"": ["*.csv", "*.json", "*.geojson"]},
    keywords=[
        "RTL",
        "Risk",
        "Tools",
        "Trading",
        "Crude",
        "Oil",
        "Refinery",
        "Refined Products",
        "Products",
    ],
    url="https://github.com/bbcho/risktools-dev",
    # download_url="https://github.com/bbcho/risktools-dev/archive/v0.5.0-beta.1.tar.gz",
    project_urls={
        # "Bug Tracker": "https://github.com/statsmodels/statsmodels/issues",
        "Documentation": "https://risktools.readthedocs.io/en/latest/",
        # "Source Code": "https://github.com/statsmodels/statsmodels",
    },
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=preqs,
    ext_package="risktools",
    cmdclass={"build_ext": NoSuffixBuilder},
    ext_modules=Cython.Build.cythonize(extensions),
)



# Performance Analytics Functions

import pandas as _pd
import numpy as _np
from sklearn.linear_model import LinearRegression as _LinearRegression


def return_cumulative(r, geometric=True):
    """
    Based on the function Return.annualize from the R package PerformanceAnalytics
    by Peter Carl and Brian G. Peterson

    Calculate a compounded (geometric) cumulative return. Based om R's PerformanceAnalytics

    This is a useful function for calculating cumulative return over a period of
    time, say a calendar year.  Can produce simple or geometric return.

    product of all the individual period returns

    .. math:: (1+r_{1})(1+r_{2})(1+r_{3})\ldots(1+r_{n})-1=prod(1+R)-1

    Parameters
    ----------
    r : Series or DataFrame
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    geometric : bool
        utilize geometric chaining (TRUE) or simple/arithmetic chaining (FALSE) to aggregate returns, by default True

    Returns
    -------
    floating point number if r is a Series, or pandas Series if r is a DataFrame

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> spy = data.DataReader("SPY",  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
    >>> spy = spy.pct_change()
    >>> rt.return_cumulative(spy['Adj Close'])
    >>> rt.return_cumulative(spy['Adj Close'], geometric=False)
    """
    r = r.dropna()

    if geometric:
        r = r.add(1).prod() - 1
    else:
        r = r.sum()

    return r


def return_annualized(r, scale=None, geometric=True):
    """
    Based on the function Return.annualize from the R package PerformanceAnalytics
    by Peter Carl and Brian G. Peterson

    Calculate an annualized return for comparing instruments with different
    length history

    An average annualized return is convenient for comparing returns.

    Annualized returns are useful for comparing two assets.  To do so, you must
    scale your observations to an annual scale by raising the compound return to
    the number of periods in a year, and taking the root to the number of total
    observations:
    
    .. math:: prod(1+R_{a})^{\\frac{scale}{n}}-1=\sqrt[n]{prod(1+R_{a})^{scale}}-1

    where scale is the number of periods in a year, and n is the total number of
    periods for which you have observations.

    For simple returns (geometric=FALSE), the formula is:

    .. math:: \overline{R_{a}} \cdot scale

    Parameters
    ----------
    r : Series or DataFrame
        pandas Series with a datetime index with freq defined. Freq must be either 'D','W','M','Q' or 'Y'
    scale : int
        number of periods in a year (daily scale = 252, monthly scale =
        12, quarterly scale = 4). By default None. Note that if scale is None,
        the function will calculate a scale based on the index frequency. If however
        you wish to override this (becuase maybe there is no index freq),
        specify your own scale to use.
    geometric : bool
        utilize geometric chaining (True) or simple/arithmetic chaining (False) to aggregate returns,
        by default True

    Returns
    -------
    floating point number if r is a Series, or pandas Series if r is a DataFrame

    References
    ----------
    Bacon, Carl. *Practical Portfolio Performance Measurement and Attribution*. Wiley. 2004. p. 6

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["SPY","AAPL"],  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
    >>> df = df.pct_change()
    >>> df = df.asfreq('B')
    >>> rt.return_annualized(r=df[('Adj Close','SPY')], geometric=True)
    >>> rt.return_annualized(r=df[('Adj Close','SPY')], geometric=False)
    >>> rt.return_annualized(r=df['Adj Close'], geometric=True)
    >>> rt.return_annualized(r=df['Adj Close'], geometric=False)
    """
    r, scale = _check_ts(r, scale, name="r")

    r = r.dropna()
    n = r.shape[0]

    if geometric:
        res = (r.add(1).cumprod() ** (scale / n) - 1).iloc[-1]
    else:
        res = r.mean() * scale
    return res


def return_excess(R, Rf=0):
    """
    Calculates the returns of an asset in excess of the given risk free rate

    Calculates the returns of an asset in excess of the given "risk free rate"
    for the period.

    Ideally, your risk free rate will be for each period you have returns
    observations, but a single average return for the period will work too.

    Mean of the period return minus the period risk free rate

    .. math:: \overline{(R_{a}-R_{f})}

    OR

    mean of the period returns minus a single numeric risk free rate

    .. math:: \overline{R_{a}}-R_{f}

    Note that while we have, in keeping with common academic usage, assumed that
    the second parameter will be a risk free rate, you may also use any other
    timeseries as the second argument.  A common alteration would be to use a
    benchmark to produce excess returns over a specific benchmark, as
    demonstrated in the examples below.

    Parameters
    ----------
    R : Series or DataFrame
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    Rf : float
        risk free rate, in same period as your returns, or as a single
        digit average

    References
    ----------
    Bacon, Carl. \emph{Practical Portfolio Performance Measurement
    and Attribution}. Wiley. 2004. p. 47-52

    Examples
    --------
    >>> data(managers)
    >>> head(Return.excess(managers[,1,drop=FALSE], managers[,10,drop=FALSE]))
    >>> head(Return.excess(managers[,1,drop=FALSE], .04/12))
    >>> head(Return.excess(managers[,1:6], managers[,10,drop=FALSE]))
    >>> head(Return.excess(managers[,1,drop=FALSE], managers[,8,drop=FALSE]))
    """

    res = R - Rf
    return res


def sd_annualized(x, scale=None, *args):
    """
    calculate a multiperiod or annualized Standard Deviation

    Standard Deviation of a set of observations \eqn{R_{a}} is given by:

    \deqn{\sigma = variance(R_{a}) , std=\sqrt{\sigma} }{std = sqrt(var(R))}

    It should follow that the variance is not a linear function of the number of
    observations.  To determine possible variance over multiple periods, it
    wouldn't make sense to multiply the single-period variance by the total
    number of periods: this could quickly lead to an absurd result where total
    variance (or risk) was greater than 100%.  It follows then that the total
    variance needs to demonstrate a decreasing period-to-period increase as the
    number of periods increases. Put another way, the increase in incremental
    variance per additional period needs to decrease with some relationship to
    the number of periods. The standard accepted practice for doing this is to
    apply the inverse square law. To normalize standard deviation across
    multiple periods, we multiply by the square root of the number of periods we
    wish to calculate over. To annualize standard deviation, we multiply by the
    square root of the number of periods per year.

    \deqn{\sqrt{\sigma}\cdot\sqrt{periods}}

    Note that any multiperiod or annualized number should be viewed with
    suspicion if the number of observations is small.

    Parameters
    ----------

    x : Series or DataFrame
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    scale : int (optional)
        number of periods in a year (daily scale = 252, monthly scale =
        12, quarterly scale = 4). By default None. Note that if scale is None,
        the function will calculate a scale based on the index frequency. If however
        you wish to override this (becuase maybe there is no index freq),
        specify your own scale to use.
    *args
        any other passthru parameters

    Returns
    -------
    floating point number if r is a Series, or pandas Series if r is a DataFrame

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["SPY","AAPL"],  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
    >>> df = df.pct_change()
    >>> df = df.asfreq('B')
    >>> rt.sd_annualized(x=df[('Adj Close','SPY')])
    >>> rt.sd_annualized(x=df['Adj Close'])
    """
    if (~isinstance(x, _pd.DataFrame) & ~isinstance(x, _pd.Series)) == True:
        raise ValueError("x must be a pandas Series or DataFrame")

    if isinstance(x.index, _pd.DatetimeIndex):
        if scale is None:
            if (x.index.freq == "D") | (x.index.freq == "B"):
                scale = 252
            elif x.index.freq == "W":
                scale = 52
            elif (x.index.freq == "M") | (x.index.freq == "MS"):
                scale = 12
            elif (x.index.freq == "Q") | (x.index.freq == "QS"):
                scale = 4
            elif (x.index.freq == "Y") | (x.index.freq == "YS"):
                scale = 1
            else:
                raise ValueError(
                    "parameter x's index must be a datetime index with freq 'D','W','M','Q' or 'Y'"
                )
    else:
        raise ValueError(
            "parameter x's index must be a datetime index with freq 'D','W','M','Q' or 'Y'"
        )

    x = x.dropna()
    res = x.std() * _np.sqrt(scale)

    return res


def omega_sharpe_ratio(R, MAR, *args):
    """
    Omega-Sharpe ratio of the return distribution

    The Omega-Sharpe ratio is a conversion of the omega ratio to a ranking statistic
    in familiar form to the Sharpe ratio.

    To calculate the Omega-Sharpe ration we subtract the target (or Minimum
    Acceptable Returns (MAR)) return from the portfolio return and we divide
    it by the opposite of the Downside Deviation.

    .. math:: OmegaSharpeRatio(R,MAR) = \\frac{R_p - R_t}{ \sum^n_{t=1} \\frac{max(R_t - R_i, 0)}{n} }

    where :math:`n` is the number of observations of the entire series

    Parameters
    ----------
    R : {Series, DataFrame}
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    MAR : {float, array_like[float]}
        Minimum Acceptable Return, in the same periodicity as your
        returns
    *args : tuple
        any other passthru parameters

    Returns
    -------
    float

    Examples
    --------
    >>> mar = 0.005
    >>> print(omega_sharpe_ratio(portfolio_bacon, MAR))
    """
    if isinstance(R, (_pd.Series, _pd.DataFrame)):
        if isinstance(R.index, _pd.DatetimeIndex) & isinstance(
            MAR, (_pd.Series, _pd.DataFrame)
        ):
            if ~isinstance(MAR.index, _pd.DatetimeIndex):
                raise ValueError(
                    "MAR index must be a datatime index if MAR and R are a Dataframe or Series with a datetime index"
                )
        elif ~isinstance(R.index, _pd.DatetimeIndex) & isinstance(
            MAR, (_pd.Series, _pd.DataFrame)
        ):
            if isinstance(MAR.index, _pd.DatetimeIndex):
                raise ValueError(
                    "R does not have a datetime index but MAR does. If both DataFrames or Series, index types must be the same"
                )

    R = R.dropna()
    r = R[R.gt(MAR)]

    if isinstance(MAR, (_pd.Series, _pd.DataFrame)) & isinstance(
        R, (_pd.Series, _pd.DataFrame)
    ):
        # subset to the same dates as the R data. Already checked that if both are series or dataframes, that
        # indices are of the same type
        MAR = MAR[r.index]
    else:
        # works for any array_like MAR. Scalars will just return itself
        # if MAR is array_like, we have to assume that R and MAR both
        # cover the same time period
        MAR = _np.mean(MAR)

    result = (
        upside_risk(R, MAR, stat="potential")
        - downside_deviation(R, MAR, potential=True)
    ) / downside_deviation(R, MAR, potential=True)
    return result


def upside_risk(R, MAR=0, method="full", stat="risk"):
    """
    upside risk, variance and potential of the return distribution

    Upside Risk is the similar of semideviation taking the return above the
    Minimum Acceptable Return instead of using the mean return or zero
    .
    To calculate it, we take the subset of returns that are more than the target
    (or Minimum Acceptable Returns (MAR)) returns and take the differences of
    those to the target.  We sum the squares and divide by the total number of
    returns and return the square root.

    Parameters
    ----------
    R : {Series, DataFrame}
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    MAR : int
        Minimum Acceptable Return, in the same periodicity as your
        returns
    method : {'full','subset'}, default 'full'
        Either "full" or "subset", indicating whether to use the
        length of the full series or the length of the subset of the series below
        the MAR as the denominator, defaults to "full"
    stat : {"risk", "variance", "potential"}, default 'risk'
        one of "risk", "variance" or "potential" indicating whether
        to return the Upside risk, variance or potential. By default,
        'risk'

    Returns
    -------
        float

    Notes
    -----
    Equations:

    .. math:: 
        
        UpsideRisk(R, MAR) = \sqrt{\sum_{n}^{t=1} \\frac{max[(R_{t} - MAR), 0]^2}{n}}\f
        UpsideVariance(R, MAR) = \sum^{n}_{t=1} \\frac{max[(R_{t} - MAR), 0]^2}{n}\f
        UpsidePotential(R, MAR) = \sum^{n}_{t=1} \\frac{max[(R_{t} - MAR), 0]} {n}

    where `n` is either the number of observations of the entire series or
    the number of observations in the subset of the series falling below the
    MAR.

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["SPY","AAPL"],  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
    >>> df = df.pct_change()
    >>> df = df.asfreq('B')
    >>> print(rt.upside_risk(df['Adj Close'], MAR=0))
    >>> print(rt.upside_risk(df[('Adj Close','SPY')], MAR=0))
    >>> print(rt.upside_risk(df['Adj Close'], MAR=0, stat='variance'))
    >>> print(rt.upside_risk(df['Adj Close'], MAR=0, stat='potential'))
    """
    # .. math:: UpsideRisk(R , MAR) = \sqrt{\sum^{n}_{t=1} \frac{max[(R_{t} - MAR), 0]^2}{n}}
    # .. math:: UpsideVariance(R, MAR) = \sum^{n}_{t=1} \frac{max[(R_{t} - MAR), 0]^2} {n}}
    # .. math:: UpsidePotential(R, MAR) = \sum^{n}_{t=1} \frac{max[(R_{t} - MAR), 0]} {n}}

    if isinstance(R, (_pd.Series, _pd.DataFrame)):
        if isinstance(R.index, _pd.DatetimeIndex) & isinstance(
            MAR, (_pd.Series, _pd.DataFrame)
        ):
            if ~isinstance(MAR.index, _pd.DatetimeIndex):
                raise ValueError(
                    "MAR index must be a datatime index if MAR and R are a Dataframe or Series with a datetime index"
                )
        elif ~isinstance(R.index, _pd.DatetimeIndex) & isinstance(
            MAR, (_pd.Series, _pd.DataFrame)
        ):
            if isinstance(MAR.index, _pd.DatetimeIndex):
                raise ValueError(
                    "R does not have a datetime index but MAR does. If both DataFrames or Series, index types must be the same"
                )

    R = R.dropna()
    r = R[R.gt(MAR)]

    if isinstance(MAR, (_pd.Series, _pd.DataFrame)) & isinstance(
        R, (_pd.Series, _pd.DataFrame)
    ):
        # subset to the same dates as the R data. Already checked that if both are series or dataframes, that
        # indices are of the same type
        MAR = MAR[r.index]
    else:
        # works for any array_like MAR. Scalars will just return itself
        # if MAR is array_like, we have to assume that R and MAR both
        # cover the same time period
        MAR = _np.mean(MAR)

    if method == "full":
        length = len(R)
    else:
        length = len(r)

    if stat == "risk":
        result = _np.sqrt(r.sub(MAR).pow(2).div(length).sum())
    elif stat == "variance":
        result = r.sub(MAR).pow(2).div(length).sum()
    else:
        result = r.sub(MAR).div(length).sum()

    return result


def downside_deviation(R, MAR=0, method="full", potential=False):
    """
    Downside deviation, similar to semi deviation, eliminates positive returns
    when calculating risk.  To calculate it, we take the returns that are less
    than the target (or Minimum Acceptable Returns (MAR)) returns and take the
    differences of those to the target.  We sum the squares and divide by the
    total number of returns to get a below-target semi-variance.

    This is also useful for calculating semi-deviation by setting
    MAR = mean(R)

    Parameters
    ----------
    R : Series or DataFrame
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    MAR : int
        Minimum Acceptable Return, in the same periodicity as your
        returns
    method : str
        Either "full" or "subset", indicating whether to use the
        length of the full series or the length of the subset of the series below
        the MAR as the denominator, defaults to "full"
    potential : bool
        potential if True, calculate downside potential instead, by default False
    *args :
        any other passthru parameters

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["SPY","AAPL"],  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
    >>> df = df.pct_change()
    >>> df = df.asfreq('B')
    >>> print(rt.downside_deviation(df[('Adj Close','SPY')], MAR=0))
    >>> print(rt.downside_deviation(df['Adj Close'], MAR=0))
    """

    if isinstance(R, (_pd.Series, _pd.DataFrame)):
        if isinstance(R.index, _pd.DatetimeIndex) & isinstance(
            MAR, (_pd.Series, _pd.DataFrame)
        ):
            if ~isinstance(MAR.index, _pd.DatetimeIndex):
                raise ValueError(
                    "MAR index must be a datatime index if MAR and R are a Dataframe or Series with a datetime index"
                )
        elif ~isinstance(R.index, _pd.DatetimeIndex) & isinstance(
            MAR, (_pd.Series, _pd.DataFrame)
        ):
            if isinstance(MAR.index, _pd.DatetimeIndex):
                raise ValueError(
                    "R does not have a datetime index but MAR does. If both DataFrames or Series, index types must be the same"
                )

    R = R.dropna()
    r = R[R.lt(MAR)]

    if isinstance(MAR, (_pd.Series, _pd.DataFrame)) & isinstance(
        R, (_pd.Series, _pd.DataFrame)
    ):
        # subset to the same dates as the R data. Already checked that if both are series or dataframes, that
        # indices are of the same type
        MAR = MAR[r.index]
    else:
        # works for any array_like MAR. Scalars will just return itself
        # if MAR is array_like, we have to assume that R and MAR both
        # cover the same time period
        MAR = _np.mean(MAR)

    if method == "full":
        length = len(R)
    else:
        length = len(r)

    if potential:
        result = r.mul(-1).add(MAR).div(length).sum()
    else:
        result = _np.sqrt(r.mul(-1).add(MAR).pow(2).div(length).sum())
        # result = r.mul(-1).add(MAR).pow(2).div(length).sum().apply(_np.sqrt)

    return result


def sharpe_ratio_annualized(R, Rf=0, scale=None, geometric=True):
    """
    calculate annualized Sharpe Ratio

    The Sharpe Ratio is a risk-adjusted measure of return that uses standard
    deviation to represent risk.

    The Sharpe ratio is simply the return per unit of risk (represented by
    variance).  The higher the Sharpe ratio, the better the combined performance
    of "risk" and return.

    This function annualizes the number based on the scale parameter.

    .. math:: \\frac{\sqrt[n]{prod(1+R_{a})^{scale}}-1}{\sqrt{scale}\cdot\sqrt{\sigma}}

    Using an annualized Sharpe Ratio is useful for comparison of multiple return
    streams.  The annualized Sharpe ratio is computed by dividing the annualized
    mean monthly excess return by the annualized monthly standard deviation of
    excess return.

    William Sharpe now recommends Information Ratio preferentially to the
    original Sharpe Ratio.

    Parameters
    ----------
    R : array_like
        an xts, vector, matrix, data frame, timeSeries or zoo object of
        asset returns
    Rf : float
        risk free rate, in same period as your returns. By default 0
    scale : int
        number of periods in a year (daily scale = 252, monthly scale =
        12, quarterly scale = 4). By default None
    geometric : bool
        utilize geometric chaining (True) or simple/arithmetic chaining (False) to aggregate returns,
        default True

    see also \code{\link{SharpeRatio}} \cr \code{\link{InformationRatio}} \cr
    \code{\link{TrackingError}} \cr \code{\link{ActivePremium}} \cr
    \code{\link{SortinoRatio}}

    References
    ----------
    Sharpe, W.F. The Sharpe Ratio,\emph{Journal of Portfolio
    Management},Fall 1994, 49-58.

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["SPY","AAPL"],  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
    >>> df = df.pct_change()
    >>> df = df.asfreq('B')
    >>> rt.sharpe_ratio_annualized(df, Rf=0, scale=None, geometric=True)
    """
    R, scale = _check_ts(R, scale)

    xR = return_excess(R, Rf)
    res = return_annualized(xR, scale=scale, geometric=geometric)
    res /= sd_annualized(R, scale=scale)

    return res


def drawdowns(R, geometric=True):
    """
    Function to calculate drawdown levels in a timeseries

    Parameters
    ----------
    R : {Series, DataFrame}
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    geometric : bool
        utilize geometric chaining (TRUE) or simple/arithmetic chaining (FALSE) to aggregate returns, by default True
    """

    if geometric:
        res = R.add(1).cumprod()
    else:
        res = R.cumsum() + 1

    res = res / res.clip(lower=1).cummax() - 1

    return res


def find_drawdowns(R, geometric=True, *args):
    """
    Find the drawdowns and drawdown levels in a timeseries.

    find_drawdowns() will find the starting period, the ending period, and
    the amount and length of the drawdown.

    Often used with sort_drawdowns() to get the largest drawdowns.

    drawdowns() will calculate the drawdown levels as percentages, for use
    in \code{\link{chart.Drawdown}}.

    Parameters
    ----------
    R : Series or DataFrame
        pandas Series or DataFrame with a datetime index. If DataFrame, function will calculate
        cummmulative returns on each column
    geometric : bool
        utilize geometric chaining (TRUE) or simple/arithmetic chaining (FALSE) to aggregate returns, by default True

    Returns
    -------
    Nested dictionary with asset name(s) as the top level key(s). Nested below that as another
    dictionary are the following keys and values:

    'return'- numpy array of minimum of returns below the risk free rate of return (Rf) for each
        trough. If returns are positive, array has 0 value
    'from' - array index positiion of beginning of each trough or recovery period corresponding to each
        element of 'return'
    'trough' - array index position of peak trough period corresponding to each
        element of 'return'. Returns beginning of recovery periods
    'to' - array index positiion of end of each trough or recovery period corresponding to each
        element of 'return'
    'length' - length of each trough period corresponding to each element of 'return' as given by
        the difference in to and from index positions
    'peaktotrough' - array index distance from the peak of each trough or recovery period from the
        beginning of each trough or recovery period, corresponding to each element of 'return'
    'recovery' - array index distance from the peak of each trough or recovery period to the
        end of each trough or recovery period, corresponding to each element of 'return'

    References
    ----------
    Bacon, C. *Practical Portfolio Performance Measurement and Attribution*. Wiley. 2004. p. 88

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["SPY","AAPL"],  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
    >>> df = df.pct_change()
    >>> df = df.asfreq('B')
    >>> rs = rt.find_drawdowns(df['Adj Close'])
    >>> print(rs['SPY']['peaktotrough'])
    """
    dd = drawdowns(R, geometric=geometric).dropna()

    # init result dict
    rs = dict()

    # convert series into dataframe for flexibility
    series_flag = False
    if isinstance(dd, _pd.Series):
        dd = _pd.DataFrame({"drawdown": dd})
        series_flag = True

    for lab, con in dd.items():
        rs[lab] = dict()
        rs[lab]["return"] = _np.array([]).astype(float)
        rs[lab]["from"] = _np.array([]).astype(int)
        rs[lab]["to"] = _np.array([]).astype(int)
        rs[lab]["length"] = _np.array([]).astype(int)
        rs[lab]["trough"] = _np.array([]).astype(int)

        if con[0] >= 0:
            prior_sign = 1
        else:
            prior_sign = 0

        frm = 0
        to = 0
        dmin = 0
        sofar = con[0]

        for i, r in enumerate(con):  # .iteritems():
            if r < 0:
                this_sign = 0
            else:
                this_sign = 1

            if this_sign == prior_sign:
                if r < sofar:
                    sofar = r
                    dmin = i
                to = i + 1
            else:
                rs[lab]["return"] = _np.append(rs[lab]["return"], sofar)
                rs[lab]["from"] = _np.append(rs[lab]["from"], frm)
                rs[lab]["trough"] = _np.append(rs[lab]["trough"], dmin)
                rs[lab]["to"] = _np.append(rs[lab]["to"], to)

                frm = i
                sofar = r
                to = i + 1
                dmin = i
                prior_sign = this_sign

        rs[lab]["return"] = _np.append(rs[lab]["return"], sofar)
        rs[lab]["from"] = _np.append(rs[lab]["from"], frm)
        rs[lab]["trough"] = _np.append(rs[lab]["trough"], dmin)
        rs[lab]["to"] = _np.append(rs[lab]["to"], to)

        rs[lab]["length"] = rs[lab]["to"] - rs[lab]["from"] + 1
        rs[lab]["peaktotrough"] = rs[lab]["trough"] - rs[lab]["from"] + 1
        rs[lab]["recovery"] = rs[lab]["to"] - rs[lab]["trough"]

        # if original parameter was a series, remove top layer of
        # results dictionary
        if series_flag == True:
            rs = rs["drawdown"]

    return rs


def _beta(y, x, subset=None):
    """
    Function to perform a linear regression on y = f(x) -> y = beta*x + const and returns
    beta

    Parameters
    ----------
    y : array-like
        dependent variable
    x : array-like
        independent variable
    subset : array-like of type bool
        array/series/list of bool to subset x and Y

    Examples
    --------
    >>> import numpy as np
    >>> import risktools as rt
    >>> x = np.array([1,2,3,4,5,6,7,8])
    >>> y = x + 5
    >>> rt._beta(x,y)
    """
    x = x.dropna()
    y = y.dropna()

    if subset is None:
        subset = _np.repeat(True, len(x))
    else:
        subset = subset.dropna().astype(bool)

    if (
        (isinstance(x, (_np.ndarray, _pd.Series)) == False)
        & (isinstance(y, (_np.ndarray, _pd.Series)) == False)
        & (isinstance(subset, (_np.ndarray, _pd.Series)) == False)
    ):
        raise ValueError(
            "all arguements of _beta must be pandas Series or numpy arrays"
        )

    # convert to arrays
    x = _np.array(x)
    y = _np.array(y)
    subset = _np.array(subset)

    # subset
    x = x[subset]
    y = y[subset]

    model = _LinearRegression()
    model.fit(x.reshape((-1, 1)), y)
    beta = model.coef_

    return beta[0]


def CAPM_beta(Ra, Rb, Rf=0, kind="all"):
    """
    calculate single factor model (CAPM) beta

    The single factor model or CAPM Beta is the beta of an asset to the variance
    and covariance of an initial portfolio.  Used to determine diversification potential.

    This function uses a linear intercept model to achieve the same results as
    the symbolic model used by \code{\link{BetaCoVariance}}

    .. math:: \\beta_{a,b}=\\frac{CoV_{a,b}}{\sigma_{b}}=\\frac{\sum((R_{a}-\\bar{R_{a}})(R_{b}-\\bar{R_{b}}))}{\sum(R_{b}-\\bar{R_{b}})^{2}}

    Ruppert(2004) reports that this equation will give the estimated slope of
    the linear regression of :math:`R_{a}` on :math:`R_{b}` and that this
    slope can be used to determine the risk premium or excess expected return
    (see Eq. 7.9 and 7.10, p. 230-231).

    kind='bull' and kind='bear' apply the same notion of best fit to positive and
    negative market returns, separately. kind='bull' is a
    regression for only positive market returns, which can be used to understand
    the behavior of the asset or portfolio in positive or 'bull' markets.
    Alternatively, kind='bear' provides the calculation on negative
    market returns.

    The function `timing_ratio` may help assess whether the manager is a good timer
    of asset allocation decisions.  The ratio, which is calculated as
    
    :math:`TimingRatio =\\frac{\\beta^{+}}{\\beta^{-}}`
    
    is best when greater than one in a rising market and less than one in a
    falling market.

    While the classical CAPM has been almost completely discredited by the
    literature, it is an example of a simple single factor model,
    comparing an asset to any arbitrary benchmark.

    Parameters
    ----------
    Ra : {array_like | DataFrame}
        Array-like or DataFrame with datetime index of asset returns to be tested vs benchmark
    Rb : array_like
        Benchmark returns to use to test Ra
    Rf : {array_like | float}
        risk free rate, in same period as your returns, or as a single
        digit average
    kind : {'all','bear','bull'}, default 'all'
        Market type to return, by default 'all'
        'all' : returns beta for all market types
        'bear' : returns beta for bear markets
        'bull' : returns beta for bull markets

    Returns
    -------
    If Ra is array-like, function returns beta as a scalar. If Ra is a DataFrame, it will return
    a series indexed with asset names from df columns

    References
    ----------
    Sharpe, W.F. Capital Asset Prices: A theory of market equilibrium under conditions of risk. *Journal of finance*, vol 19, 1964, 425-442.\f
    Ruppert, David. *Statistics and Finance, an Introduction*. Springer. 2004.\f
    Bacon, Carl. *Practical portfolio performance measurement and attribution*. Wiley. 2004.

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["XOM","AAPL","SPY"],  "yahoo", datetime(2010,1,1), datetime(2017,12,31))
    >>> df = df.pct_change()['Adj Close']
    >>> df = df.asfreq('B')
    >>> print(rt.CAPM_beta(df[['XOM','AAPL']], df['SPY'], Rf=0, kind='bear'))
    >>> print(rt.CAPM_beta(df[['XOM','AAPL']], df['SPY'], Rf=0, kind='bull'))
    >>> print(rt.CAPM_beta(df[['XOM','AAPL']], df['SPY'], Rf=0))
    """

    xRa = return_excess(Ra, Rf)
    xRb = return_excess(Rb, Rf)

    if kind == "bear":
        subset = xRb.lt(0).mask(
            xRb.isna(), _np.nan
        )  # need to include the mask, otherwise lt/gt returns False for NaN
    elif kind == "bull":
        subset = xRb.gt(0).mask(
            xRb.isna(), _np.nan
        )  # need to include the mask, otherwise lt/gt returns False for NaN
    else:
        subset = None

    if isinstance(xRa, _pd.DataFrame):
        # applies function _beta to each columns of df
        rs = xRa.apply(lambda x: _beta(x, xRb, subset), axis=0)
    else:
        rs = _beta(xRa, xRb, subset)

    return rs


def timing_ratio(Ra, Rb, Rf=0):
    """
    The function `timing_ratio` may help assess whether the manager is a good timer
    of asset allocation decisions.  The ratio is best when greater than one in a 
    rising market and less than one in a falling market.

    Parameters
    ----------
    Ra : array-like or DataFrame
        Array-like or DataFrame with datetime index of asset returns to be tested vs benchmark
    Rb : array-like
        Benchmark returns to use to test Ra
    Rf : array-like | float
        risk free rate, in same period as your returns, or as a single
        digit average

    Returns
    -------
    If Ra is array-like, function returns beta as a scalar. If Ra is a DataFrame, it will return
    a series indexed with asset names from df columns

    Notes
    -----

    .. math:: TimingRatio = \\frac{\\beta^{+}}{\\beta^{-}}

    Examples
    --------
    >>> import risktools as rt
    >>> from pandas_datareader import data, wb
    >>> from datetime import datetime
    >>> df = data.DataReader(["XOM","AAPL","SPY"],  "yahoo", datetime(2010,1,1), datetime(2017,12,31))
    >>> df = df.pct_change()['Adj Close']
    >>> df = df.asfreq('B')
    >>> rt.timing_ratio(df[['XOM','AAPL']], df['SPY'], Rf=0)
    """

    beta_bull = CAPM_beta(Ra, Rb, Rf=Rf, kind="bull")
    beta_bear = CAPM_beta(Ra, Rb, Rf=Rf, kind="bear")

    result = beta_bull / beta_bear

    return result


def _check_ts(R, scale, name="R"):
    """
    Function to check frequency of R, a time series Series or Dataframe

    Parameters
    ----------
    R : Series or DataFrame
        A time series of asset returns
    scale : int (optional)
        number of periods in a year (daily scale = 252, monthly scale =
        12, quarterly scale = 4). By default None. Note that if scale is None,
        the function will calculate a scale based on the index frequency. If however
        you wish to override this (becuase maybe there is no index freq),
        specify your own scale to use.

    Returns
    -------
    tuple with R as Series or Dataframe and scale as int
    """
    if (~isinstance(R, _pd.DataFrame) & ~isinstance(R, _pd.Series)) == True:
        raise ValueError(f"{name} must be a pandas Series or DataFrame")

    if isinstance(R.index, _pd.DatetimeIndex):
        if scale is None:
            if (R.index.freq == "D") | (R.index.freq == "B"):
                scale = 252
            elif R.index.freq == "W":
                scale = 52
            elif (R.index.freq == "M") | (R.index.freq == "MS"):
                scale = 12
            elif (R.index.freq == "Q") | (R.index.freq == "QS"):
                scale = 4
            elif (R.index.freq == "Y") | (R.index.freq == "YS"):
                scale = 1
            else:
                raise ValueError(
                    f"parameter {name}'s index must be a datetime index with freq 'D','B','W','M','Q' or 'Y'"
                )
    else:
        raise ValueError(
            f"parameter {name}'s index must be a datetime index with freq 'D','B','W','M','Q' or 'Y'"
        )

    return R, scale




# def simOUcmp(s0, theta, mu, dt, sigma, T, sims=10, eps=None, max_processors=None):
#     ST = time.time()
#     fun = _import_csimOU()
#     N = int(T/dt)

#     if max_processors is None:
#         max_processors = int(mp.cpu_count())
    
#     pool = mp.Pool(max_processors)

#     mpsim = [int(sims/max_processors)]*max_processors
#     mpsim[-1] += (sims % mpsim[-1])

#     path_size = N + 1
    
#     try:
#         iter(mu)
#         if len(mu) != N+1:
#             raise ValueError("if mu is passed as an iterable, it must be of length int(T/dt) + 1 to account for starting value s0")
#     except:
#         mu = _np.ones((N+1))*mu

#     mu = _np.tile(_np.array(mu), sims)

#     if eps is None:
#         x = _np.random.normal(loc=0, scale=_np.sqrt(dt), size=((N+1)*sims))
#         x[0] = s0
#         x[::(N+1)] = s0
#     else:
#         x = eps.T
#         x = x.reshape((N+1)*sims)
#         x[:,0] = s0
    
#     print(time.time() - ST)
    
#     for i in range(max_processors):
#         s_idx = 0 if i == 0 else (path_size*mpsim[i-1]*i)
#         e_idx = (path_size*mpsim[i]*(i+1))

#         payload = [
#             x[s_idx:e_idx],
#             theta,
#             mu,
#             dt,
#             sigma,        
#             N+1,
#             mpsim[i]
#         ]

#         rr = pool.apply_async(
#             fun,
#             args = payload
#             # kwds=payload,
#         )

#     pool.close()
#     pool.join()
#     pool.terminate()
    
#     # fun(x, theta, mu, dt, sigma, N+1, sims)
    
#     return _pd.DataFrame(x.reshape((sims, N+1)).T)

# def simOUmp(s0=5, mu=4, theta=2, sigma=1, T=1, dt=1 / 252, sims=1000, max_processors=None):

#     result = Result()

#     if max_processors is None:
#         max_processors = int(mp.cpu_count())
#     pool = mp.Pool(max_processors)

#     mpsim = [int(sims/max_processors)]*max_processors

#     mpsim[-1] += (sims % mpsim[-1])

#     print(mpsim)

#     for i in range(max_processors):
#         print(time.time() - ST)
#         payload = dict(
#             s0=s0,
#             mu=mu,
#             theta=theta,
#             sigma=sigma,
#             T=T,
#             dt=dt,
#             sims=mpsim[i]
#         )

#         rr = pool.apply_async(
#             _simOUc,
#             #args=...
#             kwds=payload,
#             callback=result.update_result
#         )

#     pool.close()
#     pool.join()
#     pool.terminate()

#     return result.val#.T.reset_index(drop=True).T